from __future__ import annotations

import asyncio
import json
import logging
import sys
import time
from asyncio import AbstractEventLoop
from typing import TYPE_CHECKING, Any, Coroutine, Literal, Optional, TypeVar, Union

import aiohttp
from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectionError

from . import __version__
from .errors import APIOffline, HTTPException, InternalServerError, UnknownStatusCode
from .types.image import AddImageText as AddImageTextPayload
from .types.image import ImageToAscii as ImageToAsciiPayload
from .types.image import Laugh as LaughPayload
from .types.random import RandomWordData
from .types.screenshot import ScreenshotData
from .types.searching import GetSearchResultData

if TYPE_CHECKING:
    from .client import Client

    T = TypeVar("T")
    Response = Coroutine[Any, Any, T]

LOGGER = logging.getLogger("ciberedev.http")

__all__ = []


async def json_or_text(
    response: aiohttp.ClientResponse, /
) -> Union[dict[str, Any], str]:
    text = await response.text(encoding="utf-8")
    try:
        if response.headers["content-type"] == "application/json":
            try:
                return json.loads(text)
            except json.JSONDecodeError:
                pass
    except KeyError:
        pass

    return text


async def error_or_text(data: Union[dict, str]) -> str:
    if isinstance(data, dict):
        return data["error"]
    else:
        return data


class Route:
    __slots__ = ["method", "endpoint"]

    def __init__(
        self,
        *,
        method: Literal["POST", "GET"],
        endpoint: str,
    ):
        self.method = method
        self.endpoint = endpoint


class HTTPClient:
    _session: Optional[ClientSession]
    _client: Client
    _loop: Optional[AbstractEventLoop]
    latency: Optional[float]
    requests: int

    __slots__ = ["_session", "_client", "_loop", "user_agent", "latency", "requests"]

    def __init__(self, *, session: Optional[ClientSession], client: Client):
        self._session = session
        self._client = client
        self._loop: Optional[AbstractEventLoop] = None
        user_agent = "ciberedev.py (https://github.com/cibere/ciberedev.py {0}) Python/{1[0]}.{1[1]} aiohttp/{2}"
        self.user_agent = user_agent.format(
            __version__, sys.version_info, aiohttp.__version__
        )
        self.requests = 0
        self.latency = None

    async def ping(self) -> float:
        route = Route(method="GET", endpoint="https://api.cibere.dev/ping")

        before = time.perf_counter()
        await self.request(route)
        after = time.perf_counter()

        latency = after - before
        self.latency = latency
        return latency

    async def request(self, route: Route, **kwargs) -> Any:
        if self._session is None:
            self._session = ClientSession()
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        self.requests += 1

        headers = kwargs.pop("headers", {})
        headers["User-Agent"] = self.user_agent
        url = route.endpoint
        endpoint = f"/{route.method.split('/')[-1]}"

        if "json" in kwargs:
            headers["Content-Type"] = "application/json"

        for current_try in range(3):
            try:
                res = await self._session.request(
                    route.method, url, ssl=False, **kwargs
                )
                data = await json_or_text(res)
            except ClientConnectionError:
                raise APIOffline(endpoint)

            if 300 > res.status >= 200:
                return data
            elif res.status == 500:
                time = 2 * current_try

                LOGGER.warning(
                    "API returned a 500 status code at '%s'. Retrying in %s seconds",
                    endpoint,
                )
                await asyncio.sleep(time)
                continue
            elif res.status == 429:
                self._loop.create_task(self._client.on_ratelimit(endpoint))
                await asyncio.sleep(5)
                return await self.request(route)
            elif res.status == 502:
                raise APIOffline(endpoint)
            elif res.status == 400:
                error = await error_or_text(data)
                raise HTTPException(error)
            else:
                raise UnknownStatusCode(res.status)

        raise InternalServerError()

    async def get_image_from_url(self, url: str) -> bytes:
        if self._session is None:
            self._session = ClientSession()
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        res = await self._session.get(url, ssl=False)
        if res.status == 200:
            return await res.read()
        else:
            txt: str = await json_or_text(res)  # type: ignore[PylancereportGeneralTypeIssues] # very wierd error that is false
            raise HTTPException(txt)

    def take_screenshot(self, url: str, delay: int) -> Response[ScreenshotData]:
        args = {"url": url, "delay": delay}
        route = Route(
            method="POST",
            endpoint="https://api.cibere.dev/screenshot",
        )

        return self.request(route, params=args)

    def get_search_results(
        self, query: str, amount: int
    ) -> Response[GetSearchResultData]:
        args = {"query": query, "amount": amount}
        route = Route(
            method="GET",
            endpoint="https://api.cibere.dev/search",
        )

        return self.request(route, params=args)

    def get_random_words(self, amount: int) -> Response[RandomWordData]:
        args = {"amount": str(amount)}
        route = Route(
            method="GET",
            endpoint="https://api.cibere.dev/random/word",
        )
        return self.request(route, params=args)

    def convert_image_to_ascii(
        self, url: str, width: Optional[int] = None
    ) -> Response[ImageToAsciiPayload]:
        args = {"url": url}
        if width:
            args["width"] = str(width)

        route = Route(
            method="GET",
            endpoint="https://api.cibere.dev/image/ascii",
        )
        return self.request(route, params=args)

    def add_text_to_image(
        self, url: str, text: str, color: tuple[int, int, int]
    ) -> Response[AddImageTextPayload]:
        data = {"url": url, "text": text, "color": list(color)}

        route = Route(
            method="GET",
            endpoint="https://api.cibere.dev/image/add-text",
        )

        return self.request(route)

    def image_laugh(
        self,
        url: Optional[str] = None,
        bytes_: Optional[bytes] = None,
        style: Literal[1, 2] = 2,
    ) -> Response[LaughPayload]:
        data: dict[str, Any] = {"style": style}

        if url:
            data["url"] = url
        if bytes_:
            data["bytes"] = bytes_

        route = Route(method="GET", endpoint="https://api.cibere.dev/image/laugh")
        return self.request(route, json=data)

    def invert_image(
        self, url: Optional[str] = None, bytes_: Optional[bytes] = None
    ) -> Response[LaughPayload]:
        data: dict[str, Any] = {}

        if url:
            data["url"] = url
        if bytes_:
            data["bytes"] = bytes_

        route = Route(method="GET", endpoint="https://api.cibere.dev/image/invert")
        return self.request(route, json=data)



import os
import platform
import sys
import webbrowser

import aiohttp

import ciberedev

commands = []


class Command:
    def __init__(self, *, name: str, description: str, full_name: str):
        self.name = name
        self.full_name = full_name
        self.description = description

    def callback(self):
        pass


class UpdateCmd(Command):
    def __init__(self):
        super().__init__(
            name="u",
            description="updates your installation of ciberedev.py",
            full_name="update",
        )

    def callback(self):
        if ciberedev.version_info.releaselevel == "final":
            lib = "ciberedev.py"
        else:
            lib = "git+https://github.com/cibere/ciberedev.py"

        os.system(sys.executable + f" -m pip install {lib}")


class OpenDocs(Command):
    def __init__(self):
        super().__init__(
            name="d",
            description="Opens the documentation in your browser",
            full_name="open-docs",
        )

    def callback(self):
        version = (
            "stable" if ciberedev.version_info.releaselevel == "final" else "latest"
        )

        print(f"Opening the {version} docs in your browser")
        webbrowser.open(f"https://docs.cibere.dev/{version}/index")


class GetVersionCmd(Command):
    def __init__(self):
        super().__init__(
            name="v",
            full_name="version",
            description="gives you the version of ciberedev.py you are running",
        )

    def callback(self):
        print(f"ciberedev.py version: {ciberedev.__version__}")


class GetSystemInfoCmd(Command):
    def __init__(self):
        super().__init__(
            name="s",
            full_name="system-info",
            description="gives you info about your system",
        )

    def callback(self):
        info = {}

        info["python"] = "v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}".format(
            sys.version_info
        )
        info["ciberedev.py"] = "v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}".format(
            ciberedev.version_info
        )
        info["aiohttp"] = aiohttp.__version__
        info["OS"] = platform.platform()

        nl = "\n"
        print(nl.join([f"{item}: {info[item]}" for item in info]))


class HelpCmd(Command):
    def __init__(self):
        super().__init__(name="h", description="the help menu", full_name="help")

    def callback(self):
        nl = "\n"
        print(
            f"""
ciberedev.py arguments

{nl.join(
    [
        f"-{cmd.name} | {cmd.full_name}: {cmd.description}"
        for cmd in commands
    ]
)}
"""
        )


commands.append(GetSystemInfoCmd())
commands.append(GetVersionCmd())
commands.append(HelpCmd())
commands.append(UpdateCmd())
commands.append(OpenDocs())


def main():
    args = sys.argv

    if len(args) == 1:
        given_cmd = "h"
    elif len(args) > 2:
        return print("Too many arguments passed")
    elif not args[1].startswith("-"):
        return print("All Arguments must start with '-'")
    else:
        given_cmd = args[1].split("-")
        given_cmd.pop(0)
        given_cmd = "".join(given_cmd)

    found_cmds = [cmd for cmd in commands if cmd.name == given_cmd]
    if len(found_cmds) == 0:
        return print(f"Unknown command {given_cmd}'")
    else:
        cmd = found_cmds[0]

    cmd.callback()


if __name__ == "__main__":
    main()



from .types.searching import SearchResultData

__all__ = ["SearchResult"]


class SearchResult:
    title: str
    description: str
    desc: str
    url: str

    __slots__ = ["title", "description", "desc", "url"]

    def __init__(self, *, data: SearchResultData):
        """Creates a SearchResult object.

        THIS SHOULD NOT BE CREATED MANUALLY, LET THE INTERNALS CREATE THEM

        Parameters
        -----------
        data: `dict`
            The raw search result data

        Attributes
        -----------
        title: `str`
            The search results title
        description: `str`
            the search results description
        desc: `str`
            alias for `SearchResult.description`
        url: `str`
            the search results url
        """

        self.title = data["title"]
        self.description = data["description"]
        self.desc = self.description
        self.url = data["url"]

    def __repr__(self):
        return f"<SearchResult title={self.title} url={self.url}>"



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from aiohttp import ClientSession

from .errors import ClientAlreadyClosed
from .file import File
from .http import HTTPClient
from .searching import SearchResult

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = ["Client"]

LOGGER = logging.getLogger(__name__)
URL_REGEX = re.compile(
    r"^(?:http|ftp)s?://"
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    r"(?::\d+)?"
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)


class Client:
    _http: HTTPClient
    _started: bool

    __slots__ = ["_http", "_started"]

    def __init__(self, *, session: Optional[ClientSession] = None):
        """Lets you create a client instance

        Parameters
        ----------
        session: Optional[`aiohttp.ClientSession`]
            an optional aiohttp client session that the internals will use for API calls

        Attributes
        ----------
        latency: `float`
            The latency between the client and the api.
        requests: `int`
            The amount of requests sent to the api during the programs lifetime
        """

        self._http = HTTPClient(session=session, client=self)
        self._started = True

    @property
    def latency(self) -> Optional[float]:
        """The latency between the client and the api.

        This variable stores the result of the last time `ciberedev.client.Client.ping` was called. By default it is `None`
        """

        return self._http.latency

    @property
    def requests(self) -> int:
        """The amount of requests sent to the api during the programs lifetime"""

        return self._http.requests

    def is_closed(self) -> bool:
        """Returns a bool depending on if the client has been closed or not

        Returns
        ----------
        bool
            True if the client is closed, False if its not been closed
        """

        return not self._started

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self, exception_type, exception_value, exception_traceback
    ) -> None:
        await self.close()

    async def on_ratelimit(self, endpoint: str) -> None:
        """|coro|

        This function is auto triggered when it hits a rate limit.

        When overriding this, you can call the super init if you still want the library to send the logs

        Parameters
        ----------
        endpoint: `str`
            the endpoint the ratelimit was hit at. Ex: '/screenshot'
        """

        LOGGER.warning(
            f"We are being ratelimited at '{endpoint}'. Trying again in 5 seconds"
        )

    async def close(self) -> None:
        """|coro|

        Closes the aiohttp session

        Raises
        ----------
        ClientAlreadyClosed
            This is raised when you already closed the client
        """

        if not self._started:
            raise ClientAlreadyClosed()

        if self._http._session:
            await self._http._session.close()

    async def take_screenshot(self, url: str, /, *, delay: int = 0) -> File:
        """|coro|

        Takes a screenshot of the given url

        Parameters
        ----------
        url: `str`
            The url you want to be screenshotted
        delay: Optional[`int`]
            The delay between going to the website, and taking the screenshot


        Raises
        ----------
        UnableToConnect
            If the api is unable to connect to the provided website
        InvalidURL
            If the url given is invalid
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        ciberedev.file.File
            A file object of your screenshot
        """

        url = url.removeprefix("<").removesuffix(">")

        if not url.startswith("http"):
            url = f"http://{url}"
        if delay > 20:
            raise TypeError("Delay must be below 20")
        if delay < 0:
            raise TypeError("Delay can not be in the negatives")
        if not re.match(URL_REGEX, url):
            raise TypeError("Invalid URL Given")

        data = await self._http.take_screenshot(url, delay)
        link = data["link"]
        fp = await self._http.get_image_from_url(link)

        file = File(raw_bytes=fp, url=link)
        return file

    async def get_search_results(
        self, query: str, /, *, amount: int = 5
    ) -> list[SearchResult]:
        """|coro|

        Searches the web with the given query

        Parameters
        ----------
        query: `str`
            The query of your search
        amount: Optional[`int`]
            The amount of results you want. Defaults to 5

        Raises
        ----------
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        List[ciberedev.searching.SearchResult]
            A list of your search results
        """

        data = await self._http.get_search_results(query, amount)

        final = []
        for raw_result in data["results"]:
            result = SearchResult(data=raw_result)
            final.append(result)

        return final

    async def get_random_words(self, amount: int, /) -> list[str]:
        """|coro|

        Gives you random words

        Parameters
        ----------
        amount: `int`
            the amount of random words you want

        Raises
        ----------
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        List[`str`]
            the random words that have been generated
        """

        data = await self._http.get_random_words(amount)
        return data["words"]

    async def convert_image_to_ascii(
        self, url: str, /, *, width: Optional[int] = None
    ) -> str:
        """|coro|

        Converts the given image to ascii art

        Parameters
        ----------
        url: `str`
            the images url
        width: Optional[`int`]
            the ascii arts width

        Raises
        ----------
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        str
            the ascii art"""

        if not re.match(URL_REGEX, url):
            raise TypeError("Invalid URL Given")

        data = await self._http.convert_image_to_ascii(url, width)
        art = data["msg"]
        return art

    async def add_text_to_image(
        self,
        *,
        image_url: str,
        text: str,
        text_color: Optional[tuple[int, int, int]] = None,
    ) -> File:
        """|coro|

        Adds text to a given image

        Parameters
        ----------
        image_url: `str`
            the images url
        text: `str`
            the text to be added
        text_color: tuple[`int`, `int`, `int`]
            the color to be added to the text

        Raises
        ----------
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        ciberedev.file.File
            A file with the new image
        """

        color = text_color or (
            255,
            255,
            255,
        )
        if not re.match(URL_REGEX, image_url):
            raise TypeError("Invalid URL Given")
        for value in color:
            if value > 255:
                raise TypeError("Invalid color given")

        data = await self._http.add_text_to_image(image_url, text, color)
        url = data["link"]
        fp = await self._http.get_image_from_url(url)

        file = File(raw_bytes=fp, url=url)
        return file

    async def image_laugh(
        self, fp: Union[str, bytes], /, *, style: Optional[Literal[1, 2]] = None
    ) -> File:
        """|coro|

        makes an image that laughs at the given image

        Parameters
        ----------
        fp : `Union[str, bytes]`
            the url or bytes of the image
        style : `Literal[1, 2]`
            the style of laugh

        Raises
        ----------
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        ciberedev.file.File
            A file with the new image
        """

        kwargs: dict[str, Any] = {"style": style or 2}

        if isinstance(fp, bytes):
            kwargs["bytes"] = fp
        else:
            kwargs["url"] = fp

        data = await self._http.image_laugh(**kwargs)
        url = data["link"]
        fp = await self._http.get_image_from_url(url)

        file = File(raw_bytes=fp, url=url)
        return file

    async def invert_image(self, fp: Union[str, bytes], /) -> File:
        """|coro|

        makes an image that laughs at the given image

        Parameters
        ----------
        fp : `Union[str, bytes]`
            the url or bytes of the image

        Raises
        ----------
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        ciberedev.file.File
            A file with the new image
        """

        kwargs: dict[str, Any] = {}

        if isinstance(fp, bytes):
            kwargs["bytes"] = fp
        else:
            kwargs["url"] = fp

        data = await self._http.image_laugh(**kwargs)
        url = data["link"]
        fp = await self._http.get_image_from_url(url)

        file = File(raw_bytes=fp, url=url)
        return file

    async def ping(self) -> float:
        """|coro|

        Pings the api

        Raises
        ----------
        UnknownError
            The api has returned an unknown error
        APIOffline
            I could not connect to the api

        Returns
        ----------
        float
            the latency. Multiply by 1000 to convert to ms
        """

        return await self._http.ping()



from typing import TypedDict


class SearchResultData(TypedDict):
    title: str
    description: str
    url: str


class GetSearchResultData(TypedDict):
    results: list[SearchResultData]
    status_code: int



"""
cibere.dev python wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper cibere.dev

.. include:: ../extras/index_page.md
.. include:: ../extras/update-log.md
"""

__description__ = "A basic wrapper cibere.dev"
__version__ = "0.5.1"

from typing import Literal, NamedTuple

from .client import *
from .errors import *
from .file import *
from .searching import *


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str


raw_version = __version__.split(".")
try:
    raw_release_releaselevel = {"a": "alpha", "b": "beta", "c": "candidate"}.get(
        raw_version[2][1], "final"
    )
except IndexError:
    raw_release_releaselevel = "final"
version_info = VersionInfo(
    major=int(raw_version[0]),
    minor=int(raw_version[1]),
    micro=int(raw_version[2][0]),
    releaselevel=raw_release_releaselevel,
)

del Literal, NamedTuple, VersionInfo, raw_release_releaselevel, raw_version



import asyncio
from functools import partial as partial_func
from typing import Any, Optional

__all__ = ["File"]


def _write_to_file(filepath: str, data: Any) -> None:
    with open(filepath, "wb") as f:
        f.write(data)


class File:
    def __init__(self, *, raw_bytes: bytes, url: Optional[str] = None):
        """Creates a file object

        Parameters
        ----------
        raw_bytes: `bytes`
            The bytes of the file
        url: Optional[`url`]
            the files url (if it has one)

        Attributes
        ----------
        bytes: `bytes`
            The bytes of the file
        url: Optional[`url`]
            the files url (if it has one)
        """

        self.bytes = raw_bytes
        self.url = url

    async def save(self, fp: str, /) -> None:
        """Saves the file

        Paramters
        ----------
        fp: `str`
            The filepath/filename the file should be saved to
        """

        func = partial_func(_write_to_file, fp, self.bytes)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, func)



import setuptools

import ciberedev

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

with open("requirements.txt", "r") as f:
    REQUIREMENTS = f.read().splitlines()

packages = ["ciberedev", "ciberedev.types"]

setuptools.setup(
    name="ciberedev.py",
    author="cibere",
    author_email="cibere.dev@gmail.com",
    url="https://github.com/cibere/ciberedev.py",
    project_urls={
        "Code": "https://github.com/cibere/ciberedev.py",
        "Issue tracker": "https://github.com/cibere/ciberedev.py/issues",
        "Discord/Support Server": "https://discord.gg/2MRrJvP42N",
    },
    version=ciberedev.__version__,
    python_requires=">=3.8",
    install_requires=REQUIREMENTS,
    packages=packages,
    description=ciberedev.__description__,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="MIT",
)



from typing import TypedDict


class BaseImage(TypedDict):
    link: str
    status_code: int


class ImageToAscii(TypedDict):
    msg: str
    status_code: int


class AddImageText(BaseImage):
    pass


class Laugh(BaseImage):
    pass



__all__ = [
    "ClientAlreadyClosed",
    "APIOffline",
    "UnknownDataReturned",
    "HTTPException",
    "UnknownStatusCode",
    "InternalServerError",
]


class CiberedevException(Exception):
    pass


class HTTPException(CiberedevException):
    def __init__(self, error: str):
        """Creates an APIException error stinace

        This is raised when the api returns an error
        It is not recommended to raise this yourself

        """

        super().__init__(error)


class UnknownStatusCode(HTTPException):
    def __init__(self, code: int):
        """Creates an UnknownStatusCode error stinace

        This is raised when the api returns an unknown status code
        It is not recommended to raise this yourself

        Parameters
        ----------
        code: `int`
            the status code that was returned

        Attributes
        ----------
        code: `int`
            the status code that was returned
        """

        self.code = code
        super().__init__(f"API returned an unknown status code: '{self.code}'")


class UnknownDataReturned(HTTPException):
    def __init__(self, endpoint: str):
        """Creates an UnknownDataReturned error stinace

        This is raised when the data the api returns does not match what the client believes it should return
        It is not recommended to raise this yourself

        Parameters
        ----------
        endpoint: `str`
            The endpoint the client is making a request to when this happend

        Attributes
        ----------
        endpoint: `str`
            The endpoint the client is making a request to when this happend
        """

        self.endpoint = endpoint
        super().__init__(
            f"API returned unknown data when making a request to '{endpoint}'"
        )


class APIOffline(HTTPException):
    def __init__(self, endpoint: str):
        """Creates an APIOffline error instance.

        This is raised when the client can not connect to the api
        It is not recommended to raise this yourself

        Parameters
        ----------
        endpoint: `str`
            the endpoint the client is trying to make a request to

        Attributes
        ----------
        endpoint: `str`
            the endpoint the client is trying to make a request to
        """

        self.endpoint = endpoint
        super().__init__(f"API is down. Aborting API request to '{endpoint}'")


class ClientAlreadyClosed(CiberedevException):
    def __init__(self):
        """Creates a ClientAlreadyClosed error instance.

        It is not recommended to raise this yourself
        """

        super().__init__(
            "Client has not been started. You can start it with 'client.run' or 'client.start'"
        )


class InternalServerError(CiberedevException):
    def __init__(self):
        """Creates a InternalServerError error instance

        It is not recommended to raise this yourself"""
        super().__init__("We have run out of attempts, aborting request")
